/**
  @author Shin'ichiro Nakaoka
*/

#include "SimpleControllerItem.h"
#include "SimpleController.h"
#include <cnoid/Link>
#include <cnoid/Sensor>
#include <cnoid/Archive>
#include <cnoid/MessageView>
#include <cnoid/ExecutablePath>
#include <cnoid/FileUtil>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/make_shared.hpp>
#include "gettext.h"

using namespace std;
using namespace boost;
using namespace boost::lambda;
using namespace cnoid;


SimpleControllerItem::SimpleControllerItem()
{
    controller = 0;
    doReloading = true;
    mv = MessageView::instance();
}


SimpleControllerItem::SimpleControllerItem(const SimpleControllerItem& org)
    : ControllerItem(org)
{
    controllerDllName = org.controllerDllName;
    doReloading = org.doReloading;
    controller = 0;
    mv = MessageView::instance();
}


SimpleControllerItem::~SimpleControllerItem()
{
    unloadController();
}


void SimpleControllerItem::unloadController()
{
    if(controller){
        delete controller;
        controller = 0;
    }

    if(controllerDll.unload()){
        mv->putln(fmt(_("The DLL of %1% \"%2%\" has been unloaded."))
                  % name() % controllerDllFileName);
    }
}


void SimpleControllerItem::onDisconnectedFromRoot()
{
    if(!isActive()){
        unloadController();
    }
}


ItemPtr SimpleControllerItem::doDuplicate() const
{
    return new SimpleControllerItem(*this);
}


void SimpleControllerItem::setControllerDllName(const std::string& name)
{
    unloadController();
    controllerDllName = name;
    controllerDllFileName.clear();
}


bool SimpleControllerItem::initialize(Body* body, double worldTimeStep)
{
    bool result = false;

    if(!controller){

        filesystem::path dllPath(controllerDllName);
        if(!checkAbsolute(dllPath)){
            dllPath =
                filesystem::path(executableTopDirectory()) /
                CNOID_PLUGIN_SUBDIR / "simplecontroller" / dllPath;
        }
        controllerDllFileName = getNativePathString(dllPath);
        controllerDll.setFileName(controllerDllFileName.c_str());
        
        if(controllerDll.isLoaded()){
            mv->putln(fmt(_("The DLL of %1% has already been loaded.")) % name());
            
            // This should be called to make the reference to the DLL.
            // Otherwise, QLibrary::unload() unloads the DLL without considering this instance.
            controllerDll.load();
            
        } else {
            mv->put(fmt(_("Loading the DLL of %1%: \"%2%\" ... ")) % name() % controllerDllFileName);
            if(!controllerDll.load()){
                mv->put(_("Failed.\n"));
                mv->putln(controllerDll.errorString());
            } else {                
                mv->putln(_("OK!"));
            }
        }
        
        if(controllerDll.isLoaded()){
            SimpleController::Factory factory =
                (SimpleController::Factory)controllerDll.resolve("createSimpleController");
            if(!factory){
                mv->putln(_("The factory function \"createSimpleController()\" is not found in the controller DLL."));
            } else {
                controller = factory();
                if(!controller){
                    mv->putln(_("The factory failed to create a controller instance."));
                } else {
                    mv->putln(_("A controller instance has successfully been created."));
                }
            }
        }
    }

    if(controller){
        timeStep_ = worldTimeStep;
        internalBody = body->duplicate();
        result = controller->initialize(internalBody.get(), timeStep_, mv->cout());
        if(!result){
            mv->putln(fmt(_("%1%'s initialize method failed.")) % name());
            if(doReloading){
                finalize();
            }
        } else {
            simulationBody = body;
        }
    }

    return result;
}


double SimpleControllerItem::timeStep()
{
    return timeStep_;
}


void SimpleControllerItem::input()
{
    for(int i=0; i < simulationBody->numJoints(); ++i){
        Link* joint0 = simulationBody->joint(i);
        Link* joint1 = internalBody->joint(i);
        joint1->q = joint0->q;
        joint1->dq = joint0->dq;
        joint1->u = joint0->u;
    }

    // copy sensor values
    int ns = simulationBody->numSensors(Sensor::FORCE);
    for(int i=0; i < ns; ++i){
        internalBody->sensor<ForceSensor>(i)->copyValuesFrom(*simulationBody->sensor<ForceSensor>(i));
    }
    ns = simulationBody->numSensors(Sensor::RATE_GYRO);
    for(int i=0; i < ns; ++i){
        internalBody->sensor<RateGyroSensor>(i)->copyValuesFrom(*simulationBody->sensor<RateGyroSensor>(i));
    }
    ns = simulationBody->numSensors(Sensor::ACCELERATION);
    for(int i=0; i < ns; ++i){
        internalBody->sensor<AccelSensor>(i)->copyValuesFrom(*simulationBody->sensor<AccelSensor>(i));
    }
}


bool SimpleControllerItem::control()
{
    return controller->control(internalBody.get(), mvout());
}


void SimpleControllerItem::output()
{
    for(int i=0; i < simulationBody->numJoints(); ++i){
        simulationBody->joint(i)->u = internalBody->joint(i)->u;
    }
}

    
void SimpleControllerItem::finalize()
{
    if(doReloading || !findRootItem()){
        unloadController();
    }
}


bool SimpleControllerItem::onReloadingChanged(bool on)
{
    doReloading = on;
    return true;
}


void SimpleControllerItem::doPutProperties(PutPropertyFunction& putProperty)
{
    putProperty(_("Controller DLL"), controllerDllName,
                (bind(&SimpleControllerItem::setControllerDllName, this, _1), true));
    putProperty(_("Reloading"), doReloading,
                (bind(&SimpleControllerItem::onReloadingChanged, this, _1)));
}


bool SimpleControllerItem::store(Archive& archive)
{
    if(archive.writeRelocatablePath("controller", controllerDllName)){
        // need to distinguish just a controller name and a controller path
        archive.write("isControllerPath", true); 
    }
    archive.write("reloading", doReloading);
    return true;
}


bool SimpleControllerItem::restore(const Archive& archive)
{
    if(archive.get("isControllerPath", false)){
        archive.readRelocatablePath("controller", controllerDllName);
    } else {
        archive.read("controller", controllerDllName);
    }
    archive.read("reloading", doReloading);
    return true;
}
